﻿namespace TestProject
{
    /// <summary>
    /// 분수
    /// </summary>
    public class Fraction
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 분자
        /// </summary>
        public long Numerator;
        
        /// <summary>
        /// 분모
        /// </summary>
        public long Denominator;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Fraction(text)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="text">텍스트</param>
        public Fraction(string text)
        {
            string[] valueArray = text.Split('/');

            Numerator   = long.Parse(valueArray[0]);
            Denominator = long.Parse(valueArray[1]);

            Simplify();
        }

        #endregion
        #region 생성자 - Fraction(numerator, denominator)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="numerator">분자</param>
        /// <param name="denominator">분모</param>
        public Fraction(long numerator, long denominator)
        {
            Numerator   = numerator;
            Denominator = denominator;

            Simplify();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region - 연산자 재정의하기 - -(a)

        /// <summary>
        /// - 연산자 재정의하기
        /// </summary>
        /// <param name="a">분수 A</param>
        /// <returns>분수</returns>
        public static Fraction operator -(Fraction a)
        {
            return new Fraction(-a.Numerator, a.Denominator);
        }

        #endregion
        #region + 연산자 재정의하기 - +(a, b)

        /// <summary>
        /// + 연산자 재정의하기
        /// </summary>
        /// <param name="a">분수 A</param>
        /// <param name="b">분수 B</param>
        /// <returns>분수</returns>
        public static Fraction operator +(Fraction a, Fraction b)
        {
            long gcd = MathHelper.GetGreatestCommonDivisor(a.Denominator, b.Denominator);

            long numerator   = a.Numerator * (b.Denominator / gcd) + b.Numerator * (a.Denominator / gcd);
            long denominator = a.Denominator * (b.Denominator / gcd);

            return new Fraction(numerator, denominator);
        }

        #endregion
        #region - 연산자 재정의하기 - -(a, b)

        /// <summary>
        /// - 연산자 재정의하기
        /// </summary>
        /// <param name="a">분수 A</param>
        /// <param name="b">분수 B</param>
        /// <returns>분수</returns>
        public static Fraction operator -(Fraction a, Fraction b)
        {
            return a + -b;
        }

        #endregion
        #region * 연산자 재정의하기 - *(a, b)

        /// <summary>
        /// * 연산자 재정의하기
        /// </summary>
        /// <param name="a">분수 A</param>
        /// <param name="b">분수 B</param>
        /// <returns>분수</returns>
        public static Fraction operator *(Fraction a, Fraction b)
        {
            Fraction result1 = new Fraction(a.Numerator, b.Denominator);
            Fraction result2 = new Fraction(b.Numerator, a.Denominator);

            return new Fraction
            (
                result1.Numerator   * result2.Numerator,
                result1.Denominator * result2.Denominator
            );
        }

        #endregion
        #region / 연산자 재정의하기 - /(a, b)

        /// <summary>
        /// / 연산자 재정의하기
        /// </summary>
        /// <param name="a">분수 A</param>
        /// <param name="b">분수 B</param>
        /// <returns>분수</returns>
        public static Fraction operator /(Fraction a, Fraction b)
        {
            return a * new Fraction(b.Denominator, b.Numerator);
        }

        #endregion

        #region 실수 암시적 변환 정의하기 - double(a)

        /// <summary>
        /// 실수 암시적 변환 정의하기
        /// </summary>
        /// <param name="a">분수 A</param>
        public static implicit operator double(Fraction a)
        {
            return (double)a.Numerator / a.Denominator;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 문자열 구하기 - ToString()

        /// <summary>
        /// 문자열 구하기
        /// </summary>
        /// <returns>문자열</returns>
        public override string ToString()
        {
            return Numerator.ToString() + "/" + Denominator.ToString();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 단순화 하기 - Simplify()

        /// <summary>
        /// 단순화 하기
        /// </summary>
        private void Simplify()
        {
            if(Denominator < 0)
            {
                Numerator   = -Numerator;
                Denominator = -Denominator;
            }

            long gcd = MathHelper.GetGreatestCommonDivisor(Numerator, Denominator);

            Numerator   = Numerator   / gcd;
            Denominator = Denominator / gcd;
        }

        #endregion
    }
}